@@ -3,24 +3,29 @@ from __future__ import annotations
 import typing
 import warnings
 
-from urwid.monitored_list import MonitoredFocusList, MonitoredList
+from urwid.split_repr import remove_defaults
 
 from .columns import Columns
 from .constants import Align, Sizing, WHSettings
 from .container import WidgetContainerListContentsMixin, WidgetContainerMixin
 from .divider import Divider
+from .monitored_list import MonitoredFocusList, MonitoredList
 from .padding import Padding
 from .pile import Pile
-from .widget import Widget, WidgetError, WidgetWrap
+from .widget import Widget, WidgetError, WidgetWarning, WidgetWrap
 
 if typing.TYPE_CHECKING:
-    from collections.abc import Iterable, Sequence
+    from collections.abc import Iterable, Iterator, Sequence
 
     from typing_extensions import Literal
 
 
 class GridFlowError(WidgetError):
-    pass
+    """GridFlow specific error."""
+
+
+class GridFlowWarning(WidgetWarning):
+    """GridFlow specific warning."""
 
 
 class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListContentsMixin):
@@ -75,12 +80,45 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
         self._cache_maxcol = self._get_maxcol(())
         super().__init__(self.generate_display_widget((self._cache_maxcol,)))
 
+    def _repr_words(self) -> list[str]:
+        if len(self.contents) > 1:
+            contents_string = f"({len(self.contents)} items)"
+        elif self.contents:
+            contents_string = "(1 item)"
+        else:
+            contents_string = "()"
+        return [*super()._repr_words(), contents_string]
+
+    def _repr_attrs(self) -> dict[str, typing.Any]:
+        attrs = {
+            **super()._repr_attrs(),
+            "cell_width": self.cell_width,
+            "h_sep": self.h_sep,
+            "v_sep": self.v_sep,
+            "align": self.align,
+            "focus": self.focus_position if len(self._contents) > 1 else None,
+        }
+        return remove_defaults(attrs, GridFlow.__init__)
+
+    def __rich_repr__(self) -> Iterator[tuple[str | None, typing.Any] | typing.Any]:
+        yield "cells", [widget for widget, _ in self.contents]
+        yield "cell_width", self.cell_width
+        yield "h_sep", self.h_sep
+        yield "v_sep", self.v_sep
+        yield "align", self.align
+        yield "focus", self.focus_position
+
+    def __len__(self) -> int:
+        return len(self._contents)
+
     def _invalidate(self) -> None:
         self._cache_maxcol = None
         super()._invalidate()
 
     def _contents_modified(
-        self, slc, new_items: Iterable[tuple[Widget, tuple[Literal["given", WHSettings.GIVEN], int]]]
+        self,
+        _slc: tuple[int, int, int],
+        new_items: Iterable[tuple[Widget, tuple[Literal["given", WHSettings.GIVEN], int]]],
     ) -> None:
         for item in new_items:
             try:
@@ -123,7 +161,7 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
         )
         focus_position = self.focus_position
         self.contents = [(new, (WHSettings.GIVEN, self._cell_width)) for new in widgets]
-        if focus_position < len(widgets):
+        if focus_position < len(widgets):  # pylint: disable=consider-using-max-builtin  # pylint bug
             self.focus_position = focus_position
 
     def _get_cells(self):
@@ -255,7 +293,7 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
             return None
         return self.contents[self.focus_position][0]
 
-    def _get_focus(self) -> Widget:
+    def _get_focus(self) -> Widget | None:
         warnings.warn(
             f"method `{self.__class__.__name__}._get_focus` is deprecated, "
             f"please use `{self.__class__.__name__}.focus` property",
@@ -384,6 +422,12 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
     def _get_maxcol(self, size: tuple[int] | tuple[()]) -> int:
         if size:
             (maxcol,) = size
+            if maxcol < self.cell_width:
+                warnings.warn(
+                    f"Size is smaller than cell width ({maxcol!r} < {self.cell_width!r})",
+                    GridFlowWarning,
+                    stacklevel=3,
+                )
         else:
             maxcol = len(self) * self.cell_width + (len(self) - 1) * self.h_sep
         return maxcol
@@ -435,18 +479,15 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
                 pad.first_position = i
                 p.contents.append((pad, p.options()))
 
-            c.contents.append((w, c.options(WHSettings.GIVEN, width_amount)))
+            # Use width == maxcol in case of maxcol < width amount
+            # Columns will use empty widget in case of GIVEN width > maxcol
+            c.contents.append((w, c.options(WHSettings.GIVEN, min(width_amount, maxcol))))
             if (i == self.focus_position) or (not column_focused and w.selectable()):
                 c.focus_position = len(c.contents) - 1
                 column_focused = True
             if i == self.focus_position:
                 p.focus_position = len(p.contents) - 1
             used_space = sum(x[1][1] for x in c.contents) + self.h_sep * len(c.contents)
-            if width_amount > maxcol:
-                # special case: display is too narrow for the given
-                # width so we remove the Columns for better behaviour
-                # FIXME: determine why this is necessary
-                pad.original_widget = w
             pad.width = used_space - self.h_sep
 
         if self.v_sep:
@@ -482,7 +523,11 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
         # pad.first_position was set by generate_display_widget() above
         self.focus_position = pile_focus.first_position + col_focus_position
 
-    def keypress(self, size: tuple[int], key: str) -> str | None:
+    def keypress(
+        self,
+        size: tuple[int],  # type: ignore[override]
+        key: str,
+    ) -> str | None:
         """
         Pass keypress to display widget for handling.
         Captures focus changes.
@@ -493,7 +538,11 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
             self._set_focus_from_display_widget()
         return key
 
-    def pack(self, size: tuple[int] | tuple[()] = (), focus: bool = False) -> tuple[int, int]:
+    def pack(
+        self,
+        size: tuple[int] | tuple[()] = (),  # type: ignore[override]
+        focus: bool = False,
+    ) -> tuple[int, int]:
         if size:
             return super().pack(size, focus)
         cols = len(self) * self.cell_width + (len(self) - 1) * self.h_sep
@@ -503,7 +552,11 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
         self.get_display_widget(size)
         return super().rows(size, focus=focus)
 
-    def render(self, size: tuple[int] | tuple[()], focus: bool = False):
+    def render(
+        self,
+        size: tuple[int] | tuple[()],  # type: ignore[override]
+        focus: bool = False,
+    ):
         self.get_display_widget(size)
         return super().render(size, focus)
 
@@ -521,7 +574,7 @@ class GridFlow(WidgetWrap[Pile], WidgetContainerMixin, WidgetContainerListConten
 
     def mouse_event(
         self,
-        size: tuple[int] | tuple[()],
+        size: tuple[int] | tuple[()],  # type: ignore[override]
         event: str,
         button: int,
         col: int,
